import warnings
from collections.abc import Sequence
from functools import wraps
from pathlib import Path
from typing import Any, Literal, SupportsIndex

import numpy as np
from numpy.typing import ArrayLike
from pydantic import TypeAdapter

from ..core_defs import LCN_EX
from ..utils import _mask
from .frame_defs import FF, FT
from .frame_defs import FrameHeader as FH
from .types import FRAME_DTYPE, FrameArrayLike, FrameArrayType


class FrameIllegalError(ValueError):
    """Frame is illegal."""

    pass


class ShapeError(ValueError):
    """Exception for incorrect shape."""

    pass


class TruncationWarning(UserWarning):
    """Value out of range & will be truncated."""

    pass


OUT_OF_RANGE_WARNING = "{0} out of range, will be truncated into {1} bits, {2}."


def header2type(header: FH) -> FT:
    if header <= FH.CONFIG_TYPE4:
        return FT.CONFIG
    elif header <= FH.TEST_TYPE4:
        return FT.TEST
    elif header <= FH.WORK_TYPE4:
        return FT.WORK
    elif header <= FH.CTRL_TYPE4:
        return FT.CTRL

    raise FrameIllegalError(f"unknown header: {header}.")


def single_frame_header_check(frame: FRAME_DTYPE, expected: FH) -> bool:
    # Both for v1 & v2
    return (int(frame) >> FF.GENERAL_HEADER_OFFSET) & FF.GENERAL_HEADER_MASK == expected


def framearray_header_check(
    frames: FrameArrayType, expected_type: FH, strict: bool = True
) -> bool:
    """Check the header of frame arrays."""
    header0 = FH((int(frames[0]) >> FF.GENERAL_HEADER_OFFSET) & FF.GENERAL_HEADER_MASK)

    if header0 != expected_type:
        if strict:
            raise ValueError(
                f"expected frame type {expected_type.name}, but got {header0.name}."
            )
        else:
            return False

    headers = (frames >> FF.GENERAL_HEADER_OFFSET) & FF.GENERAL_HEADER_MASK

    if np.unique(headers).size > 1:
        if strict:
            raise ValueError(
                "the headers of the frame are not all the same, please check the frames value."
            )
        else:
            return False

    return True


def frame_array2np(frame_array: FrameArrayLike) -> FrameArrayType:
    if isinstance(frame_array, int):
        return np.asarray([frame_array], dtype=FRAME_DTYPE)

    elif isinstance(frame_array, np.ndarray):
        if frame_array.ndim != 1:
            warnings.warn(
                f"ndim of frame arrays must be 1, but got {frame_array.ndim}. Flatten anyway.",
                UserWarning,
            )
        return frame_array.flatten().astype(FRAME_DTYPE)

    elif isinstance(frame_array, (list, tuple)):
        return np.asarray(frame_array, dtype=FRAME_DTYPE)

    else:
        raise TypeError(
            f"expected int, list, tuple or np.ndarray, but got {type(frame_array).__name__}."
        )


# Frame field widths for formatting
_FRAME_COMMON_WIDTHS = [4, 5, 5, 5, 5, 5, 5]
OFF_FRAME_GENERAL_WIDTHS = _FRAME_COMMON_WIDTHS + [30]
OFF_FRAME_WORK1_WIDTHS = _FRAME_COMMON_WIDTHS + [3, 11, 8, 8]
ON_FRAME_WORK1_1_WIDTHS = _FRAME_COMMON_WIDTHS + [3, 11, 5, 3, 8]

# For chip v2.5
LCN_TO_TS_AXON_WIDTHS = (
    (8, 9),
    (7, 10),
    (6, 11),
    (5, 12),
    (4, 13),
    (3, 14),
    (2, 15),
    (1, 16),
)
_FRAME_COMMON_WIDTHS_V2 = [4, 6, 6, 6, 6, 6, 6]
OFF_FRAME_GENERAL_WIDTHS_V2 = _FRAME_COMMON_WIDTHS_V2 + [24]
OFF_FRAME_WORK1_WIDTHS_V2 = [3, 1, 6, 6, 6, 6, 6, 6] + [7, 9, 8]


def format_frame_bin(
    value: SupportsIndex,
    widths: Sequence[int] = OFF_FRAME_GENERAL_WIDTHS,
    sep: str = "_",
    reverse: bool = False,
) -> str:
    total_bits = FF.FRAME_LENGTH
    bin_str = np.binary_repr(value, width=total_bits)
    parts = []
    start = 0

    for w in reversed(widths) if reverse else widths:
        parts.append(bin_str[start : start + w])
        start += w

    return sep.join(parts)


def print_frame(
    frames: ArrayLike,
    widths: Sequence[int] | None = None,
    version: Literal[1, 2] = 1,
    *,
    sep: str = "_",
    reverse: bool = False,
    target_lcn: LCN_EX | None = None,
) -> list[str]:
    if version == 1:
        widths = widths or OFF_FRAME_GENERAL_WIDTHS
    else:
        if widths is None:
            if target_lcn is None:
                widths = OFF_FRAME_GENERAL_WIDTHS_V2
            else:
                TS_WIDTH, AX_WIDTH = LCN_TO_TS_AXON_WIDTHS[target_lcn.value]
                widths = OFF_FRAME_WORK1_WIDTHS_V2.copy()
                widths[-3] = TS_WIDTH - 1
                widths[-2] = AX_WIDTH
        else:
            widths = widths

    s = [
        format_frame_bin(f, widths, sep, reverse)
        for f in np.asarray(frames, FRAME_DTYPE).flat
    ]
    for line in s:
        print(line)

    return s


def np2npy(fp: Path, d: np.ndarray) -> None:
    assert fp.suffix == ".npy"
    np.save(fp, d)


def np2bin(fp: Path, d: np.ndarray) -> None:
    assert fp.suffix == ".bin"
    d.tofile(fp)


def np2txt(fp: Path, d: np.ndarray) -> None:
    assert fp.suffix == ".txt"

    with fp.open("w") as f:
        for i in range(d.size):
            f.write(f"{d[i]:0{FF.FRAME_LENGTH}b}\n")


_HighBit = int
_LowBit = int


def bin_split(
    x: int, pos: int, high_mask_bit: int | None = None
) -> tuple[_HighBit, _LowBit]:
    """Split an integer and return the high & low bits.

    Argument:
        - x: the integer.
        - pos: the position (LSB) to split the binary.
        - high_mask: mask for the high part. Optional.

    Example::

        >>> bin_split(0b1100001001, 3)
        97(0b1100001), 1
    """
    low = x & _mask(pos)

    if isinstance(high_mask_bit, int):
        high = (x >> pos) & _mask(high_mask_bit)
    else:
        high = x >> pos

    return high, low


def params_check(checker: TypeAdapter):
    def inner(func):
        @wraps(func)
        def wrapper(params: dict[str, Any], *args, **kwargs):
            validated = checker.validate_python(params).model_dump()  # return dict
            return func(validated, *args, **kwargs)

        return wrapper

    return inner


def params_check2(checker1: TypeAdapter, checker2: TypeAdapter):
    def inner(func):
        @wraps(func)
        def wrapper(params1: dict[str, Any], params2: dict[str, Any], *args, **kwargs):
            checked1 = checker1.validate_python(params1)
            checked2 = checker2.validate_python(params2)
            return func(checked1, checked2, *args, **kwargs)

        return wrapper

    return inner



import random
import string
from collections.abc import Iterable
from typing import TypeVar

import numpy as np


def _mask(mask_bit: int) -> int:
    return (1 << mask_bit) - 1


IT = TypeVar("IT", bound=Iterable)


def range_check(param: IT, field: str, min_val: int, max_val: int) -> IT:
    if any(item > max_val or item < min_val for item in param):
        raise ValueError(f"parameter '{field}' out of range [{min_val}, {max_val}].")

    return param


def range_check_unsigned(param: IT, field: str, max_val: int) -> IT:
    return range_check(param, field, 0, max_val)


def gen_random_string(length: int = 8) -> str:
    chars = string.ascii_letters + string.digits
    return "".join(random.choice(chars) for _ in range(length))


def ndarray_serializer(value: int | np.ndarray) -> int | list[int]:
    return value if isinstance(value, int) else value.tolist()



from enum import IntEnum, unique

from .utils import _mask

__all__ = [
    # Neuron reg limits
    "OfflineNeuRegLimV2",
    # Types
    "OutputType",
    "FoldType",
    "NeuronType",
    "ThresholdNegMode",
    "ThresholdPosMode",
    "LateralInhibitionMode",
    "LeakMultiComparisonOrder",
    "LeakMultiInputMode",
    "LeakMultiMode",
    "LeakAddMode",
    "WeightCompressType",
]


class OfflineNeuRegLimV2:
    """Limits of offline neuron registers for chip v2.5."""

    TICK_RELATIVE_MAX = _mask(8)
    ADDR_AXON_MAX = _mask(9)
    ADDR_CORE_COORD_MIN = -_mask(5)
    ADDR_CORE_COORD_MAX = _mask(5)
    WEIGHT_SKEW_MAX = _mask(16)
    WEIGHT_ADDRESS_MAX = _mask(12)
    RESET_V_MAX = _mask(15)
    RESET_V_MIN = -(RESET_V_MAX + 1)
    THRES_NEG_MAX = _mask(31)
    THRES_NEG_MIN = -(THRES_NEG_MAX + 1)
    THRES_POS_MAX = THRES_NEG_MAX
    THRES_POS_MIN = THRES_NEG_MIN
    LEAK_TAU_MAX = _mask(5)
    LEAK_TAU_MIN = -(LEAK_TAU_MAX + 1)
    LEAK_V_MAX = _mask(19)
    LEAK_V_MIN = -(LEAK_V_MAX + 1)
    VJT_INITIAL_MAX = _mask(11)
    VJT_INITIAL_MIN = -(VJT_INITIAL_MAX + 1)
    FOLD_RANGE_MAX = _mask(11)
    FOLD_SKEW_MAX = _mask(11)
    FOLD_AXON_MAX = _mask(11)
    FOLD_NUMBER_MAX = _mask(29)


@unique
class OutputType(IntEnum):
    """Output type selection.
    0: Output spike or activation value
    1: Output membrane potential
    """

    VALUE = 0
    POTENTIAL = 1


@unique
class FoldType(IntEnum):
    """Fold type selection.
    0: Unfolded neuron
    1: Folded neuron
    """

    UNFOLDED = 0
    FOLDED = 1


@unique
class NeuronType(IntEnum):
    """Neuron type selection.
    0: Half neuron mode, occupies 128 bits
    1: Full neuron mode, occupies 256 bits
    """

    HALF = 0
    FULL = 1


@unique
class ThresholdNegMode(IntEnum):
    """Negative threshold mode selection.
    0: Fire mode
    1: Floor mode
    """

    FIRE = 0
    FLOOR = 1


@unique
class ThresholdPosMode(IntEnum):
    """Positive threshold mode selection.
    0: Fire mode
    1: Ceiling mode
    """

    FIRE = 0
    CEILING = 1


@unique
class LateralInhibitionMode(IntEnum):
    """Lateral inhibition mode selection.
    0: No lateral inhibition
    1: With lateral inhibition, if the current core generated a spike in the last time step, membrane potential accumulates from reset_v
    """

    DISABLE = 0
    ENABLE = 1


@unique
class LeakMultiComparisonOrder(IntEnum):
    """Multiplicative leak sequence.
    0: Leak before threshold comparison
    1: Leak after threshold comparison reset
    """

    BEFORE_COMPARE = 0
    AFTER_COMPARE = 1


@unique
class LeakMultiInputMode(IntEnum):
    """Whether input participates in multiplicative leak.
    0: Input does not participate in multiplicative leak
    1: Input participates in multiplicative leak
    """

    DISABLE = 0
    ENABLE = 1


@unique
class LeakMultiMode(IntEnum):
    """Multiplicative leak mode selection.
    0: No multiplicative leak, but membrane potential can be shifted
    1: With multiplicative leak
    """

    DISABLE = 0
    ENABLE = 1


@unique
class LeakAddMode(IntEnum):
    """Additive leak mode selection.
    0: Forward leak mode
    1: Backward leak mode
    """

    FORWARD = 0
    BACKWARD = 1


@unique
class WeightCompressType(IntEnum):
    """Weight type.
    0: Dense mode, weight data uncompressed
    1: Sparse mode, weight data CSC compressed
    """

    DENSE = 0
    SPARSE = 1



from typing import Literal

__all__ = [
    "HwParams",
    "HwCoreParams",
    "HwOfflineCoreParams",
    "HwOnlineCoreParams",
    # For chip v2.5
    "HwParamsV2",
]


class HwParams:
    """Basic hardware configuration of the chip."""

    COORD_Y_PRIORITY: bool = True
    """Coordinate priority"""

    WEIGHT_BITORDER: Literal["little", "big"] = "little"
    N_CHIP_MAX = 1024

    N_BIT_COORD_ADDR = 5  # Address for X/Y chip/core/core* coordinates
    COORD_IS_SIGNED: bool = False
    CORE_X_MIN = 0
    CORE_X_MAX = (1 << N_BIT_COORD_ADDR) - 1
    CORE_Y_MIN = CORE_X_MIN
    CORE_Y_MAX = CORE_X_MAX

    N_CORE_MAX_INCHIP = 1024
    N_CORE_OFFLINE = 1008
    N_CORE_ONLINE = N_CORE_MAX_INCHIP - N_CORE_OFFLINE

    CORE_X_OFFLINE_MIN = CORE_X_MIN
    CORE_Y_OFFLINE_MIN = CORE_Y_MIN
    CORE_X_OFFLINE_MAX = CORE_X_MAX
    CORE_Y_OFFLINE_MAX = CORE_Y_MAX
    # NOTE: The following values are for the online cores
    CORE_X_ONLINE_MIN = 0b11100  # 28
    CORE_Y_ONLINE_MIN = 0b11100  # 28
    CORE_X_ONLINE_MAX = CORE_X_MAX  # 31
    CORE_Y_ONLINE_MAX = CORE_Y_MAX  # 31

    N_ROUTING_PATH_LENGTH_MAX = 5
    N_SUB_ROUTING_NODE = 4
    """The number of sub-level routing nodes of a routing node."""


class HwCoreParams:
    """Hardware configuration of a core."""


class HwOfflineCoreParams(HwCoreParams):
    """Hardware configuration of an offline core."""

    CORE_X_MIN = HwParams.CORE_X_OFFLINE_MIN
    CORE_Y_MIN = HwParams.CORE_Y_OFFLINE_MIN
    CORE_X_MAX = HwParams.CORE_X_OFFLINE_MAX
    CORE_Y_MAX = HwParams.CORE_Y_OFFLINE_MAX

    N_BIT_TIMESLOT = 8
    N_TIMESLOT_MAX = 1 << N_BIT_TIMESLOT

    N_FANIN_PER_DENDRITE_MAX = 1152
    N_FANIN_PER_DENDRITE_SNN = N_FANIN_PER_DENDRITE_MAX
    N_FANIN_PER_DENDRITE_ANN = 144  # in 8-bit
    """The #N of fan-in per dendrite."""

    N_DENDRITE_MAX_SNN = 512
    N_DENDRITE_MAX_ANN = 4096
    """The maximum #N of dendrites in one core."""

    N_NEURON_MAX_SNN = 512
    N_NEURON_MAX_ANN = 1888
    """The maximum #N of neurons in one core."""

    ADDR_RAM_MAX = N_NEURON_MAX_SNN - 1
    """The maximum RAM address (starting from 0)."""

    ADDR_AXON_MAX = N_FANIN_PER_DENDRITE_MAX - 1
    """The maximum axons address (starting from 0)."""

    WEIGHT_RAM_SHAPE = (512, 1152)

    FANOUT_IW8: list[int] = [N_NEURON_MAX_ANN, 1364, 876, 512, 256, 128, 64, 32, 16, 8]
    """The fan-out of 8-bit input width under different combination rate of dendrites(LCN + WW)."""


class HwOnlineCoreParams(HwCoreParams):
    """Hardware configuration of an online core."""

    CORE_X_MIN = HwParams.CORE_X_ONLINE_MIN
    CORE_Y_MIN = HwParams.CORE_Y_ONLINE_MIN
    CORE_X_MAX = HwParams.CORE_X_ONLINE_MAX
    CORE_Y_MAX = HwParams.CORE_Y_ONLINE_MAX

    N_BIT_TIMESLOT = 3
    N_TIMESLOT_MAX = 1 << N_BIT_TIMESLOT

    N_FANIN_PER_DENDRITE_MAX = 1024
    """The #N of fan-in per dendrite."""
    N_DENDRITE_MAX = 1024
    N_NEURON_MAX = N_DENDRITE_MAX

    ADDR_RAM_MAX = N_NEURON_MAX - 1
    """The maximum RAM address (starting from 0)."""

    ADDR_AXON_MAX = N_FANIN_PER_DENDRITE_MAX - 1
    """The maximum axons address (starting from 0)."""

    WEIGHT_RAM_SHAPE = (8192, 128)
    LUT_LEN = 60
    """Length of LUT on the online cores."""


class HwParamsV2:
    """Hardware configuration of the chip for v2.5."""

    N_CORE_MAX_INCHIP = 81  # 9*9
    N_CORE_OFFLINE = 63  # 7*9
    N_CORE_ONLINE = 17
    N_CORE_RV_CPU = 1

    # Use 'Z' instead of 'XY' for clearity
    COORDZXY_PRIORITY: Literal["Z", "X", "Y"] = "Z"
    COORDXY_PRIORITY: Literal["X", "Y"] = "X"
    COORD_PRIORITY = COORDXY_PRIORITY

    WEIGHT_BITORDER: Literal["little", "big"] = "little"

    # Original code form of signed numbers
    N_BIT_COORD_ADDR = 6
    COORD_IS_SIGNED: bool = True

    CORE_Z_MIN = 1 - (1 << (N_BIT_COORD_ADDR - 1))
    CORE_Z_MAX = -CORE_Z_MIN
    CORE_X_MIN = CORE_Z_MIN
    CORE_X_MAX = CORE_Z_MAX
    CORE_Y_MIN = CORE_Z_MIN
    CORE_Y_MAX = CORE_Z_MAX

    COPY_Z_MIN = CORE_Z_MIN
    COPY_Z_MAX = CORE_Z_MAX
    COPY_X_MIN = CORE_X_MIN
    COPY_X_MAX = CORE_X_MAX
    COPY_Y_MIN = CORE_Y_MIN
    COPY_Y_MAX = CORE_Y_MAX
